#include <cmath>
#include <iomanip>
#include <iostream>
#include <sstream>

inline void check_cuda_error(const char* cudaop, const char* source,
       unsigned int line, cudaError_t error) {
   if (error != cudaSuccess) {
      std::cerr << cudaop << " at " << source << ":" << line
	 << " failed: " << cudaGetErrorString(error) << std::endl;
      exit(1);
   }
}

#define CHECK_CUDA(opname, ...) \
   check_cuda_error(#opname, __FILE__, __LINE__, opname(__VA_ARGS__))

// to be integrated function
__device__ double f(double x) {
   return 4 / (1 + x*x);
}

// must be a power of 2
#define THREADS_PER_BLOCK 256

// numerical integration according to the Simpson rule
// for f over the i-th subinterval of [a,b]
__global__ void simpson(double a, double b, double* sums) {
   /* compute approximative sum for our sub-interval */
   const int N = blockDim.x * gridDim.x;
   const unsigned int i = threadIdx.x + blockIdx.x * blockDim.x;
   double xleft = a + (b - a) / N * i;
   double xright = xleft + (b - a) / N;
   double xmid = (xleft + xright) / 2;
   double sum = (xright - xleft) / 6 * (f(xleft) + 4 * f(xmid) + f(xright));

   /* store it into the per-block shared array of sums */
   unsigned int me = threadIdx.x;
   __shared__ double sums_per_block[THREADS_PER_BLOCK];
   sums_per_block[me] = sum;

   /* aggregate sums within a block */
   int index = blockDim.x / 2;
   while (index) {
      __syncthreads();
      if (me < index) {
	 sums_per_block[me] += sums_per_block[me + index];
      }
      index /= 2;
   }
   /* publish result */
   if (me == 0) {
      sums[blockIdx.x] = sums_per_block[0];
   }
}

char* cmdname;
void usage() {
   std::cerr << "Usage: " << cmdname << " [# intervals]" << std::endl;
   exit(1);
}

int main(int argc, char** argv) {
   const double a = 0;
   const double b = 1;
   unsigned int N = 8192;

   cmdname = *argv++; --argc;
   if (argc > 0) {
      std::istringstream arg(*argv++); --argc;
      if (!(arg >> N) || N <= 0) usage();
   }
   if (argc > 0) usage();

   int blocksize = THREADS_PER_BLOCK;
   if (N % blocksize != 0) {
      std::cerr << cmdname << ": please select a multiple of "
	 << blocksize << std::endl;
      exit(1);
   }
   int nof_blocks = N / blocksize;

   double* cuda_sums;
   CHECK_CUDA(cudaMalloc, (void**)&cuda_sums, N * sizeof(double));
   simpson<<<nof_blocks, blocksize>>>(a, b, cuda_sums);

   double sums[nof_blocks];
   CHECK_CUDA(cudaMemcpy, sums, cuda_sums, nof_blocks * sizeof(double), cudaMemcpyDeviceToHost);
   CHECK_CUDA(cudaFree, cuda_sums);
   double sum = 0;
   for (int i = 0; i < nof_blocks; ++i) {
      sum += sums[i];
   }
   std::cout << std::setprecision(14) << sum << std::endl;
   std::cout << std::setprecision(14) << M_PI << std::endl;
}
